/*
 *  tcfe: 3Com Fast Ethernet MAC driver for Solaris
 *  @(#)tcfe_gem.c	1.4 07/02/10
 *
 * Copyright (c) 2006-2007 Masayuki Murayama.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * 3. Neither the name of the author nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/*
 Change log
	2006/12/24 multicast hash calculation fixed
	2007/02/10 cleanup
	2007/02/10 2.4.0
 */

/*
 TODO:
 */

/*
 * System Header files.
 */
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/debug.h>
#include <sys/stropts.h>
#include <sys/stream.h>
#include <sys/strlog.h>
#include <sys/kmem.h>
#include <sys/stat.h>
#include <sys/kstat.h>
#include <sys/vtrace.h>
#include <sys/dlpi.h>
#include <sys/strsun.h>
#include <sys/ethernet.h>
#include <sys/modctl.h>
#include <sys/errno.h>
#include <sys/dditypes.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include <sys/pci.h>
#include "mii.h"
#include "gem.h"
#include "3c90xreg.h"

char	ident[] = "3com 3c90x driver v" VERSION;

/* Debugging support */
#ifdef DEBUG_LEVEL
static int tcfe_debug = DEBUG_LEVEL;
#define	DPRINTF(n, args)	if (tcfe_debug>(n)) cmn_err args
#else
#define	DPRINTF(n, args)
#endif

/*
 * Useful macros and typedefs
 */
#define	ROUNDUP(x, a)	(((x) + (a) - 1) & ~((a) - 1))

#if defined(sparc) || defined(sparcv9)
#define	LE32(x)		ddi_swap32(x)
#else
#define	LE32(x)		(x)
#endif

#define	EXEC_CMD(dp, cmd)	OUTW((dp), (WxCommand), (cmd))
#define	EXEC_CMD_WAIT(dp, cmd)	tcfe_exec_cmd_wait((dp), (cmd))

#define	SET_WIN(dp, n)		EXEC_CMD(dp, CmdSelectRegisterWindow | (n))

#define	REG_LOCK(lp)		mutex_enter(&lp->reglock)
#define	REG_UNLOCK(lp)		mutex_exit(&lp->reglock)

#define	DN_LIST_SIZE_SHIFT	6
#define	DN_LIST_SIZE		(1 << DN_LIST_SIZE_SHIFT)
#define	MAXTXFRAGS	\
	((DN_LIST_SIZE - sizeof(uint32_t) * 2) / sizeof(struct DnFrag))

#define	UP_LIST_SIZE_SHIFT	5
#define	UP_LIST_SIZE		(1 << UP_LIST_SIZE_SHIFT)
#define	MAXRXFRAGS	\
	((UP_LIST_SIZE - sizeof(uint32_t) * 2) / sizeof(struct UpFrag))

/*
 * Our configuration
 */
#define	OUR_INTR_BITS \
	(IS_hostError | IS_txComplete | IS_intRequested | \
	 IS_updateStats | IS_upComplete | IS_dnComplete | IS_interruptLatch)

#define	TXFIFOSIZE	(2*1024)
#define	RXFIFOSIZE	(2*1024)

#ifndef TX_RING_SIZE
#  define TX_RING_SIZE	256
#endif
#ifndef RX_RING_SIZE
#  define RX_RING_SIZE	256
#endif

#ifndef TX_BUF_SIZE
#  define TX_BUF_SIZE	TX_RING_SIZE
#endif
#ifndef RX_BUF_SIZE
#  define RX_BUF_SIZE	RX_RING_SIZE
#endif

#define	ONESEC		(drv_usectohz(1*1000000))

static int	tcfe_tx_copy_thresh = 256;
static int	tcfe_rx_copy_thresh = 256;

/*
 * Supported chips
 */
struct chip_info {
	uint16_t	venid;
	uint16_t	devid;
	char		*name;
	uint_t		flags;
#define	CHIP_TYPE	0x0007
#define	CHIP_TYPE_59X	0x0000
#define	CHIP_TYPE_90X	0x0001
#define	CHIP_TYPE_90XB	0x0002
#define	CHIP_TYPE_90XC	0x0003
#define	EEADDRSIZE_8	0x4000
#define	CHIP_CARDBUS	0x8000
	uint16_t	ee_offset;
	uint16_t	reset_opt;
};

#define	IS_3C59X(lp)	(((lp)->chip->flags & CHIP_TYPE) == CHIP_TYPE_59X)
#define	IS_3C90X(lp)	(((lp)->chip->flags & CHIP_TYPE) == CHIP_TYPE_90X)
#define	IS_3C90XB(lp)	(((lp)->chip->flags & CHIP_TYPE) == CHIP_TYPE_90XB)
#define	IS_3C90XC(lp)	(((lp)->chip->flags & CHIP_TYPE) == CHIP_TYPE_90XC)
#define	IS_3C90XB_OR_LATER(lp)	\
			(((lp)->chip->flags & CHIP_TYPE) >= CHIP_TYPE_90XB)
#define	IS_CARDBUS(lp)	(((lp)->chip->flags & CHIP_CARDBUS) != 0)

#define	ASIC_40_0502_00X(lp)	(IS_3C90XB(lp) && ((lp)->rev_id >> 5) == 0)
#define	ASIC_40_0483_00X(lp)	(IS_3C90XB(lp) && ((lp)->rev_id >> 5) == 1)
#define	ASIC_40_0476_001(lp)	(IS_3C90XB(lp) && ((lp)->rev_id >> 5) == 3)

static struct chip_info tcfe_chiptbl[] = {
#ifdef CONFIG_3C59X
 {	0x10B7, 0x5900, "3C590", CHIP_TYPE_59X, 0, 0, },
 {	0x10B7, 0x5920, "3C592", CHIP_TYPE_59X, 0, 0, },
 {	0x10B7, 0x5970, "3C597", CHIP_TYPE_59X, 0, 0, },
 {	0x10B7, 0x5950, "3C595", CHIP_TYPE_59X, 0, 0, },
 {	0x10B7, 0x5951, "3C595", CHIP_TYPE_59X, 0, 0, },
 {	0x10B7, 0x5952, "3C595", CHIP_TYPE_59X, 0, 0, },
#endif
 /* boomerang 10M */
 {	0x10b7, 0x9000, "3C900 10baseT", CHIP_TYPE_90X, 0, 0,},
 {	0x10b7, 0x9001, "3C900 10Mbps combo", CHIP_TYPE_90X, 0, 0,},
 /* cyclone 10M */
 {	0x10b7, 0x9004, "3C900 10Mbps TPO", CHIP_TYPE_90XB, 0, 0,},
 {	0x10b7, 0x9005, "3C900 10Mbps combo", CHIP_TYPE_90XB, 0, 0,},
 /* tornade 10M */
 {	0x10b7, 0x9006, "3C900 10Mbps TPC", CHIP_TYPE_90XB, 0, 0,},
 {	0x10b7, 0x900a, "3C900B-FL 10base-FL", CHIP_TYPE_90XB, 0, 0,},
 /* boomerang 100M */
 {	0x10b7, 0x9050, "3C905 100baseTx", CHIP_TYPE_90X, 0, 0,},
 {	0x10b7, 0x9051, "3C905 100baseT4", CHIP_TYPE_90X, 0, 0,},
 /* cyclone 100M */
 {	0x10b7, 0x9055, "3C905B 100baseTx", CHIP_TYPE_90XB, 0, 0,},
 {	0x10b7, 0x9058, "3C905B 10/100/BNC", CHIP_TYPE_90XB, 0, 0,},
 {	0x10b7, 0x905a, "3C905B-FX 100baseFx", CHIP_TYPE_90XB, 0, 0,},
 /* tornade 100M */
 {	0x10b7, 0x9200, "3C905C", CHIP_TYPE_90XC , 0, 0},
 {	0x10b7, 0x9202, "3C920B-EMB-WNM (ATI)", CHIP_TYPE_90XC , 0, 0},
 {	0x10b7, 0x9201, "3C920", CHIP_TYPE_90XC , 0, 0},
 /* 980 series cyclone */
 {	0x10b7, 0x9800, "3C980", CHIP_TYPE_90XB , 0, 0},
 {	0x10b7, 0x9805, "3C980C Python", CHIP_TYPE_90XB , 0, 0},
 /* SOHO series tornade */
 {	0x10b7, 0x7646, "3CSOHO100-TX", CHIP_TYPE_90XC, 0, 0},
 /* mini-pci 555/556w/modem series */
 {	0x10b7, 0x5055, "3C555", CHIP_TYPE_90XB | EEADDRSIZE_8, 0x30, 0},
 {	0x10b7, 0x6055, "3C556", 
		CHIP_TYPE_90XC | CHIP_CARDBUS | EEADDRSIZE_8,
			0x30, RO_invertMIIPwr},
 {	0x10b7, 0x6056, "3C556B", 
		CHIP_TYPE_90XC | CHIP_CARDBUS,
			0x30, RO_invertMIIPwr},
 /* cardbus 575 series */
 {	0x10b7, 0x5b57, "3C575", CHIP_TYPE_90X | EEADDRSIZE_8, 0x30, 0},
 {	0x10b7, 0x5057, "3C575", CHIP_TYPE_90X | EEADDRSIZE_8, 0x30, 0},
 {	0x10b7, 0x5157, "3C575BT",
		CHIP_TYPE_90XB | CHIP_CARDBUS | EEADDRSIZE_8,
			0x30, RO_invertLEDPwr},
 {	0x10b7, 0x5257, "3C575CT",
		CHIP_TYPE_90XC | CHIP_CARDBUS | EEADDRSIZE_8,
			0x30, RO_invertMIIPwr},
 /* cardbus 656 series */
 {	0x10b7, 0x6560, "3C656", 
		CHIP_TYPE_90XB | CHIP_CARDBUS | EEADDRSIZE_8,
			0x30, RO_invertMIIPwr | RO_invertLEDPwr},
 {	0x10b7, 0x6562, "3C656B", 
		CHIP_TYPE_90XB | CHIP_CARDBUS | EEADDRSIZE_8,
			0x30, RO_invertMIIPwr | RO_invertLEDPwr},
 {	0x10b7, 0x6564, "3C656C", 
		CHIP_TYPE_90XC | CHIP_CARDBUS | EEADDRSIZE_8,
			0x30, RO_invertMIIPwr},

 {	0x10b7, 0x4500, "3C450", CHIP_TYPE_90XC , 0, 0},


/* tornade chipset */
 {	0x10b7, 0x1201, "3C982", CHIP_TYPE_90XC , 0, 0},
 {	0x10b7, 0x1202, "3C982", CHIP_TYPE_90XC , 0, 0},
 {	0x10b7, 0x9056, "3C905BT4", CHIP_TYPE_90XC , 0, 0},
 {	0x10b7, 0x9210, "3C920B-EMB-WNM", CHIP_TYPE_90XC , 0, 0},
 {	0xffff, 0xffff, "unknown 3com chipset", CHIP_TYPE_90XC, 0, },
};
#define CHIPTABLESIZE   (sizeof(tcfe_chiptbl)/sizeof(struct chip_info))

struct tcfe_dev {
	uint8_t			rev_id;
	struct chip_info	*chip;

	kmutex_t		reglock;

	boolean_t		dnlistptr_loaded;
	uint32_t		intenable;

	/* cardbus staff */
	ddi_acc_handle_t	base2_ha;
	void			*base2;

	/* configuration */
	int			media;
	uint16_t		bmcr;
	uint16_t		bmsr;
	uint16_t		adv;

	/* polling mode */
#ifdef GEM_CONFIG_POLLING
	int			last_poll_interval;
#endif
	/* multicast hash */
	uint8_t			mhash[256/8];

	/* rx list mngt */
	int			rx_last_slot;

	/* global error recovery */
	boolean_t		need_to_reset;
};

/* ======================================================== */
 
/* mii operations */
static void  tcfe_mii_sync(struct gem_dev *);
static uint16_t  tcfe_mii_read(struct gem_dev *, uint_t);
static void tcfe_mii_write(struct gem_dev *, uint_t, uint16_t);

/* nic operations */
static int tcfe_reset_chip(struct gem_dev *);
static int tcfe_attach_chip(struct gem_dev *);
static int tcfe_init_chip(struct gem_dev *);
static int tcfe_start_chip(struct gem_dev *);
static int tcfe_stop_chip(struct gem_dev *);
static int tcfe_set_media(struct gem_dev *);
static int tcfe_set_rx_filter(struct gem_dev *);
static int tcfe_get_stats(struct gem_dev *);

/* descriptor operations */
static int tcfe_tx_desc_write(struct gem_dev *dp, int slot,
		    ddi_dma_cookie_t *dmacookie, int frags, uint32_t flag);
static void tcfe_rx_desc_write(struct gem_dev *dp, int slot,
		    ddi_dma_cookie_t *dmacookie, int frags);
static void tcfe_tx_start(struct gem_dev *dp, int slot, int nslot);
static void tcfe_rx_start(struct gem_dev *dp, int slot, int nslot);
static uint_t tcfe_tx_desc_stat(struct gem_dev *dp, int slot, int ndesc);
static uint64_t tcfe_rx_desc_stat(struct gem_dev *dp, int slot, int ndesc);

static void tcfe_tx_desc_init(struct gem_dev *dp, int slot);
static void tcfe_rx_desc_init(struct gem_dev *dp, int slot);
static void tcfe_tx_desc_clean(struct gem_dev *dp, int slot);
static void tcfe_rx_desc_clean(struct gem_dev *dp, int slot);

/* interrupt handler */
static u_int tcfe_interrupt(struct gem_dev *dp);

/* ======================================================== */

/* mapping attributes */
/* Data access requirements. */
static struct ddi_device_acc_attr tcfe_dev_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};

/* Tx and rx buffers should have native endianness for sparc platforms. */
static struct ddi_device_acc_attr tcfe_buf_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_NEVERSWAP_ACC,	/* native endianness */
	DDI_STRICTORDER_ACC
};

static ddi_dma_attr_t tcfe_dma_attr_buf = {
	DMA_ATTR_V0,		/* dma_attr_version */
	0,			/* dma_attr_addr_lo */
	0xffffffffull,		/* dma_attr_addr_hi */
	0x1fffull,		/* dma_attr_count_max */
	1,			/* dma_attr_align */
	0xffffffff,		/* dma_attr_burstsizes */
	1,			/* dma_attr_minxfer */
	0x1fff,			/* dma_attr_maxxfer */
	0xffffffffull,		/* dma_attr_seg */
	0,/* patched later */	/* dma_attr_sgllen */
	1,			/* dma_attr_granular */
	0			/* dma_attr_flags */
};

static ddi_dma_attr_t tcfe_dma_attr_desc = {
	DMA_ATTR_V0,		/* dma_attr_version */
	16,			/* dma_attr_addr_lo */
	0xffffffffull,		/* dma_attr_addr_hi */
	0xffffffffull,		/* dma_attr_count_max */
	16,			/* dma_attr_align */
	0xffffffff,		/* dma_attr_burstsizes */
	1,			/* dma_attr_minxfer */
	0xffffffffull,		/* dma_attr_maxxfer */
	0xffffffffull,		/* dma_attr_seg */
	1,			/* dma_attr_sgllen */
	1,			/* dma_attr_granular */
	0			/* dma_attr_flags */
};

/* ======================================================== */
/*
 * HW manupilation routines
 */
/* ======================================================== */
static void
tcfe_exec_cmd_wait(struct gem_dev *dp, int cmd)
{
	int		i;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	ASSERT(mutex_owned(&lp->reglock));

	EXEC_CMD(dp, cmd);

	for (i = 0; (INW(dp, WxIntStatus) & IS_cmdInProgress) != 0; i++) {
		if (i > 200) {
			cmn_err(CE_WARN,
				"%s: cmd timeout error: cmd:0x%04",
			       dp->name, cmd);
			lp->need_to_reset = B_TRUE;
			break;
		}
		drv_usecwait(10);
	}
}

static int
tcfe_reset_chip(struct gem_dev *dp)
{
	int		i;
	uint32_t	val;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	REG_LOCK(lp);
#ifdef NEVER
	EXEC_CMD_WAIT(dp, CmdGlobalReset | 0xff);
	delay(drv_usectohz(10000));
#endif
	if (IS_CARDBUS(lp)) {
		SET_WIN(dp, 2);
		val = INW(dp, W2ResetOptions) &
				~(RO_invertLEDPwr | RO_invertMIIPwr);
		OUTW(dp, W2ResetOptions, val | lp->chip->reset_opt);
		DPRINTF(0, (CE_CONT, "%s: %s reset option:%b",
			dp->name, __func__, val, ResetOptionsBits));
		/* enable cardbus interrupt */
		ddi_put32(lp->base2_ha,
			(uint32_t *)(((caddr_t) lp->base2) + 4), 0x8000); /*X*/

		/*
		 * XXX - need to enable tranceiver power by writing
		 * 0x0800 into reg 0 in window 0?
		 */
	}

	/* Reset transmitter */
	EXEC_CMD_WAIT(dp, CmdTxReset);
	EXEC_CMD_WAIT(dp, CmdTxReset);

	/* Reset receiver */
	EXEC_CMD_WAIT(dp, CmdRxReset);

	REG_UNLOCK(lp);

	lp->need_to_reset = B_FALSE;

	return (GEM_SUCCESS);
}

static int
tcfe_init_chip(struct gem_dev *dp)
{
	int		i;
	uint32_t	val;
	int		pktsize;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	DPRINTF(1, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	pktsize = sizeof(struct ether_header) + 4 + dp->mtu;

	REG_LOCK(lp);

	/* Set InterruptMask */
	lp->intenable = 0;
	EXEC_CMD(dp, CmdSetInterruptEnable | lp->intenable);
	EXEC_CMD(dp, CmdAcknowledgeInterrupt | AllAck);

	/* Set IndicationEnable */
	EXEC_CMD(dp, CmdSetIndicationEnable | AllIndications);

	/* setup transmitter */
	EXEC_CMD_WAIT(dp, CmdSetTxStartThresh |
		min(TxStartThreshMask, dp->txthr / TxStartThreshUnit));

	if (IS_3C90XB_OR_LATER(lp)) {
		/* XXX - use default */
		EXEC_CMD(dp, CmdSetTxReclaimThresh | 8);
	}

	/* setup rx */
	val = ROUNDUP(pktsize, RxEarlyThreshUnit) / RxEarlyThreshUnit;
	EXEC_CMD(dp, CmdSetRxEarlyThresh | min(RxEarlyThreshMask, val));

	/* common and extended registers */

	/* WxTxPktId 0x18 90xB ro */
	/* WxTimer		0x1a */
	/* WxTxStatus		0x1b */
	/* WxIntStatus_1e	0x1e */
	/* WxIntStatusAuto	0x1e */
	/* WxCommand_1e		0x1e */
	/* WxDmaCtrl		0x20 */
	val = INL(dp, WxDmaCtrl);
	val |= DC_upRxEarlyEnable;	/* rw */
	val |= DC_counterSpeed;		/* rw 0:3.2uS, 1:320nS */
	val &= ~DC_countdownMode;	/* rw 0:from now, 1:from intr*/
	if (IS_3C90XB_OR_LATER(lp)) {
		val &= ~DC_upAltSeqDisable;	/* rw 90xB/C */
		val &= ~DC_defeatMWI;		/* rw 90xB/C */
		val &= ~DC_defeatMRL;		/* rw 90xB/C */
		val &= ~DC_upOverDiscDisable;	/* rw 90xB/C */
	}
	OUTL(dp, WxDmaCtrl, val);

	/* WxDnListPtr 0x24 */
	EXEC_CMD_WAIT(dp, CmdDnStall);
	OUTL(dp, WxDnListPtr, 0);
	lp->dnlistptr_loaded = B_FALSE;

	/* WxDnBurstThresh 0x2a 90XB/C */
	if (IS_3C90XB_OR_LATER(lp)) {
		val = dp->txmaxdma/DnBurstThreshUnit;
		val = max(1, min(val, DnBurstThreshMask));
		OUTB(dp, WxDnBurstThresh, val);
	}

	/* WxDnPriorityThreash 0x2c */
	if (IS_3C90XB_OR_LATER(lp)) {
		val = TXFIFOSIZE/8/DnPriorityThreshUnit;
		val = max(1, min(val, DnPriorityThreshMask));
		OUTB(dp, WxDnPriorityThresh, val);
	}

	/* WxDnPoll 0x2d */
	if (IS_3C90XB_OR_LATER(lp)) {
#ifdef TEST_BOOMERANG
		OUTB(dp, WxDnPoll, 0);
#else
		OUTB(dp, WxDnPoll, 0x7f);
#endif
	}
	/* WxTxFreeThresh 0x2f (3c90x) */
	if (IS_3C90X(lp)) {
		/*
		 * XXX - 3C90X nics must have enough space in tx fifo
		 * when it starts DMAing a packet, otherwise it may
		 * hang frequently during downloading.
		 */
		val = ROUNDUP(pktsize, TxFreeThreshUnit) / TxFreeThreshUnit;
		OUTB(dp, WxTxFreeThresh, min(TxFreeThreshMask, val));
	}

	/* WxUpPktStatus 0x30, ro */
	/* WxFreeTimer 0x34 */
	/* WxCountdown 0x36, stop */
	OUTW(dp, WxCountdown, 0);

	/*
	 * WxUpListPtr 0x38
	 */
	EXEC_CMD_WAIT(dp, CmdUpStall);
	OUTL(dp, WxUpListPtr, (uint32_t) dp->rx_ring_dma);

	/* WxUnPriorityThreash 0x3c */
	if (IS_3C90XB_OR_LATER(lp)) {
		val = (RXFIFOSIZE/8)/UpPriorityThreshUnit;
		val = max(1, min(val, UpPriorityThreshMask));
		OUTB(dp, WxUpPriorityThresh, val);
	}

	/* WxUpPoll 0x3d */
	if (IS_3C90XB_OR_LATER(lp)) {
#ifdef TEST_BOOMERANG
		OUTB(dp, WxUpPoll, 0);
#else
		OUTB(dp, WxUpPoll, 0x7f);
#endif
	}

	/* WxUpBurstThresh 0x3e */
	if (IS_3C90XB_OR_LATER(lp)) {
		val = dp->rxmaxdma/UpBurstThreshUnit;
		val = max(1, min(val, UpBurstThreshMask));
		OUTB(dp, WxUpBurstThresh, val);
	}

	/* WxRealTimeCnt 0x40 */
	/* WxDebugData 0x70 */
	/* WxDebugControl 0x74 */
	/* WxDnMaxBurst 0x78, rw, no need to setup */
	/* WxPowerMgmtCtrl 0x7a */

	/* Window 0 */
	SET_WIN(dp, 0);
	/* W0EepromCommand:0xa, nothing */
	/* W0EepromData:0xc, nothing*/

	/* Window 1 */
	SET_WIN(dp, 1);
	/* XXX - all registers in window 1 are obsolute */

	/* Window 2 */
	SET_WIN(dp, 2);
	/* W2StationAddress 0x0, set in set_rx_filter */
	/* W2StationMask 0x6 */
	for (i = 0; i < ETHERADDRL; i += 2) {
		OUTW(dp, W2StationMask + i, 0);
	}
	/* W2ResetOptions 0xc (90xB/C) */
	/* XXX - ResetOptions register will be configured in mii_init. */

	/* Window 3 */
	SET_WIN(dp, 3);
	/* W3InternalConfig offset 0x0 dw, set in reset_chip */
	/* W3MaxPktSize 0x4 w 90XB/C */
	if (IS_3C90XB_OR_LATER(lp)) {
		OUTW(dp, W3MaxPktSize,
			sizeof(struct ether_header) + 4 + dp->mtu);
	}

	/* W3MacControl 0x6 word */
	val = INW(dp, W3MacControl);
	if (IS_3C90X(lp)) {
		val |= MC_allowLargePackets;
	}
#ifdef NEVER
	if (IS_3C90XB_OR_LATER(lp)) {
		/* we don't use 3com vlan tagging. */
		val |= MC_vltEnable;
	}
#endif
	OUTW(dp, W3MacControl, val);

	/* W3ResetOptions_old 0x8 w, obsolute, 90x */
	/* W3MediaOptions 0x8 w, will be refered in mii_init */
	/* W3RxFree 0xa	w ro */
	/* W3TxFree 0xc	w ro */

	/* Window 4 */
	SET_WIN(dp, 4);
	/* W4FifoDiagnostic 0x4	*/
	/* W4NetworkDiagnostic 0x6 */
	OUTW(dp, W4NetworkDiagnostic,
		INW(dp, W4NetworkDiagnostic) | ND_upperBytesEnable);

	/* W4PhysicalMgmt 0x8, do nothing */
	/* W4MediaStatus 0xa, configured in mii_init */
	/* W4BadSSD 0xc */
	/* W4UpperBytesOk 0xd, do nothing */

	/* Window 5 */
	SET_WIN(dp, 5);
	/* no registers to manupilate in window 5 */

	/* Window 7 */
	SET_WIN(dp, 7);
	if (IS_3C90XB_OR_LATER(lp)) {
		OUTW(dp, W7VlanEtherType, 0x8100);	/* setup TPID */
	}

	if (IS_3C90XB_OR_LATER(lp)) {
		int	hashsize = ASIC_40_0502_00X(lp) ? 64 : 256;
		/* clear entire multicast hash table */
		for (i = 0; i < hashsize; i++) {
			EXEC_CMD(dp, CmdSetHashFilterBits | i);
		}
		bzero(&lp->mhash[0], sizeof(lp->mhash));
	}

	/* end of hardware initialization */
	REG_UNLOCK(lp);

#ifdef GEM_CONFIG_POLLING
	lp->last_poll_interval = 0;
#endif
	lp->rx_last_slot = 0;

	return (GEM_SUCCESS);
}

static uint32_t
tcfe_mcast_hash(struct gem_dev *dp, uint8_t *addr)
{
	return (gem_ether_crc_be(addr, ETHERADDRL));
}

static int
tcfe_set_rx_filter(struct gem_dev *dp)	
{
	int		i;
	uint8_t		*mac;
	uint8_t		mhash[256/8];
	uint16_t	operand;
	int		hashsize;
	uint_t		h;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

#if DEBUG_LEVEL > 3
	for (i = 0; i < dp->mc_count; i++) {
		cmn_err(CE_CONT,
		"!%s: adding mcast(%d) %02x:%02x:%02x:%02x:%02x:%02x",
			dp->name, i,
			dp->mc_list[i].addr.ether_addr_octet[0],
			dp->mc_list[i].addr.ether_addr_octet[1],
			dp->mc_list[i].addr.ether_addr_octet[2],
			dp->mc_list[i].addr.ether_addr_octet[3],
			dp->mc_list[i].addr.ether_addr_octet[4],
			dp->mc_list[i].addr.ether_addr_octet[5]);
	}
#endif

	/* Normal RX operation mode */
	operand = rxFilterIndividual | rxFilterBroadcast;

	if (IS_3C90XB_OR_LATER(lp)) {
		/* we can use a 64/256bit-width hardware multicast filter */
		hashsize = ASIC_40_0502_00X(lp) ? 64 : 256;
		bzero(mhash, hashsize/8);
	}
	else {
		hashsize = 0;
	}

	if ((dp->rxmode & RXMODE_ENABLE) == 0) {
		operand = 0;
	}
	else if ((dp->rxmode & RXMODE_PROMISC) != 0) {
		/* Promiscous mode */
		operand |= rxFilterAllMulticast | rxFilterPromiscuous;
	}
	else if (dp->mc_count > hashsize/2 ||
			(dp->rxmode & RXMODE_ALLMULTI) != 0) {
		/* All Multicast mode */
		operand |= rxFilterAllMulticast;
	}
	else if (dp->mc_count > 0) {
		/*
		 * make hash table to select interresting
		 * multicast address only.
		 */
		operand |= rxFilterMulticastHash;
		for (i = 0; i < dp->mc_count; i++) {
			h = dp->mc_list[i].hash & (hashsize - 1);
			mhash[h / 8] |= 1 << (h % 8);
		}
	}

	REG_LOCK(lp);
	SET_WIN(dp, 2);
	mac = &dp->cur_addr.ether_addr_octet[0];
	for (i = 0; i < ETHERADDRL; i += 2) {
		OUTW(dp, W2StationAddress + i, (mac[i + 1] << 8) | mac[i]);
	}

	if (hashsize > 0) {
		/* update hardware multicast hash filter */
		for (i = 0; i < hashsize; i++) {
			h = (mhash[i/8] >> (i%8)) & 1;

			if (((lp->mhash[i/8] >> (i%8)) & 1) == h) {
				/* no need to change the bit in hash filter */
				continue;
			}

			EXEC_CMD(dp, CmdSetHashFilterBits | (h << 10) | i);
		}
		bcopy(&mhash[0], &lp->mhash[0], hashsize/8);
	}

	EXEC_CMD(dp, CmdSetRxFilter | operand);

	REG_UNLOCK(lp);

	return (GEM_SUCCESS);
}

static int
tcfe_set_media(struct gem_dev *dp)
{
	uint16_t	old;
	uint16_t	new;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	REG_LOCK(lp);
	SET_WIN(dp, 3);

	old = INW(dp, W3MacControl);
	new = old & ~(MC_fullDuplexEnable | MC_flowControlEnable);

	/* duplex mode */
	if (dp->full_duplex) {
		new |= MC_fullDuplexEnable;
	}

	/* flow control */
	if(dp->flow_control == FLOW_CONTROL_SYMMETRIC) {
		new |= MC_flowControlEnable;
	}

	if (new != old) {
		OUTW(dp, W3MacControl, new);
#ifdef notdef
		if (dp->mac_active) {
			/* XXX -- reset rx and tx here */
			EXEC_CMD_WAIT(dp, CmdTxReset);
			EXEC_CMD_WAIT(dp, CmdRxReset);

			drv_usecwait(2000);

			EXEC_CMD(dp, CmdTxEnable);
			EXEC_CMD(dp, CmdRxEnable);
		}
#endif
	}

	REG_UNLOCK(lp);

	return (GEM_SUCCESS);
}

static int
tcfe_start_chip(struct gem_dev *dp)
{
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	ASSERT(mutex_owned(&dp->intrlock));

	REG_LOCK(lp);

	/* Enable statistics */
	EXEC_CMD(dp, CmdStatisticsEnable);

	/* Enable interrupts */
	lp->intenable = OUR_INTR_BITS;
	EXEC_CMD(dp, CmdSetInterruptEnable | lp->intenable);

	/* Enable transmitter */
	EXEC_CMD(dp, CmdTxEnable);

	/* Enable receiver */
	EXEC_CMD(dp, CmdRxEnable);

	/* Start Dnload DMA engine */
	EXEC_CMD(dp, CmdDnUnstall);

	/* Start Upload DMA engine */
	OUTL(dp, WxUpListPtr, (uint32_t) dp->rx_ring_dma);
	EXEC_CMD(dp, CmdUpUnstall);

	REG_UNLOCK(lp);

	return (GEM_SUCCESS);
}

static int
tcfe_stop_chip(struct gem_dev *dp)
{
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	DPRINTF(1, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	REG_LOCK(lp);

	/* inhibit interrupt */
	EXEC_CMD(dp, CmdSetInterruptEnable | 0);

	/* no more statistics */
	EXEC_CMD(dp, CmdStatisticsDisable);

	/* stop receiver, then stop rx dma/	
	EXEC_CMD(dp, CmdRxDisable);
	EXEC_CMD_WAIT(dp, CmdUpStall);

	/* stop tx dma, then stop transmitter */
	EXEC_CMD_WAIT(dp, CmdDnStall);
	EXEC_CMD(dp, CmdTxDisable);

	/* wait DMA completes */
	drv_usecwait(dp->speed == GEM_SPD_100 ? dp->mtu/10 : dp->mtu);

	if (lp->media == xcvrSelect10BASE2) {
		EXEC_CMD(dp, CmdDisableDcConverter);
		drv_usecwait(800);
	}

	EXEC_CMD_WAIT(dp, CmdRxReset);
	EXEC_CMD_WAIT(dp, CmdTxReset);

	REG_UNLOCK(lp);

	return (GEM_SUCCESS);
}

static int
tcfe_get_stats(struct gem_dev *dp)
{
	int		first_coll;
	int		multi_coll;
	volatile int	x;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	DPRINTF(2, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	REG_LOCK(lp);
	EXEC_CMD(dp, CmdStatisticsDisable);

	SET_WIN(dp, 6);
	dp->stats.nocarrier 	+= INB(dp, W6CarrierLost); 
	/* heartbeat err */	x= INB(dp, W6SqeErrors);
	multi_coll		 = INB(dp, W6MultipleCollisions);
	dp->stats.multi_coll	+= multi_coll;
	first_coll		 = INB(dp, W6SingleCollisions);
	dp->stats.first_coll	+= first_coll;
	dp->stats.xmtlatecoll	+= INB(dp, W6LateCollisions);
	dp->stats.missed	+= INB(dp, W6RxOverruns);
	dp->stats.opackets	+= INB(dp, W6FramesXmittedOk);
	dp->stats.rpackets	+= INB(dp, W6FramesRcvdOk);
	dp->stats.defer  	+= INB(dp, W6FramesDeferred); 
	dp->stats.rbytes	+= INW(dp, W6BytesRcvdOk);
	dp->stats.obytes	+= INW(dp, W6BytesXmittedOk);

	/* read extended statistics */
				x= INB(dp, W6UpperFramesOk);
	dp->stats.rpackets	+=
	  ((x & UF_upperFramesRcvdOk) >> UF_upperFramesRcvdOkShift) << 8;
	dp->stats.opackets	+=
	  ((x & UF_upperFramesXmittedOk) >> UF_upperFramesXmittedOkShift) << 8;

	SET_WIN(dp, 4);
				x= INB(dp, W4BadSSD);
				x= INW(dp, W4UpperBytesOk);
	dp->stats.rpackets	+=
	((x & UF_upperFramesRcvdOk) >> UF_upperFramesRcvdOkShift) << 16;
	dp->stats.opackets	+=
	((x & UF_upperFramesXmittedOk) >> UF_upperFramesXmittedOkShift) << 16;

	/*
	 * Guess total collisions
	 */
	dp->stats.collisions += first_coll + multi_coll*2;

	EXEC_CMD(dp, CmdStatisticsEnable);

	REG_UNLOCK(lp);

	return (GEM_SUCCESS);
}

/*
 * discriptor  manupiration
 */
static int
tcfe_tx_desc_write(struct gem_dev *dp, int slot,
		ddi_dma_cookie_t *dmacookie, int frags, uint32_t flag)
{
	int			i;
	struct DPD_type0	*tfdp;
	struct DnFrag		*tfp;
	ddi_dma_cookie_t	*dcp;
	uint32_t		mark;
	uint32_t		total;
	struct tcfe_dev		*lp = (struct tcfe_dev *)dp->private;

#if DEBUG_LEVEL > 2
	/* force to cause interrupt upon tx completion */
	flag |= GEM_TXFLAG_INTR;
#endif
	tfdp = (struct DPD_type0 *)(dp->tx_ring + DN_LIST_SIZE * slot);
#if DEBUG_LEVEL > 2
	cmn_err(CE_CONT,
		"!%s: %s seqnum: %d, slot %d, frags: %d flag: %d",
		dp->name, __func__, dp->tx_desc_tail, slot, frags, flag);
	for (i = 0; i < frags; i++) {
		cmn_err(CE_CONT, "%d: addr: 0x%x, len: 0x%x",
			i, dmacookie[i].dmac_address, dmacookie[i].dmac_size);
	}
#endif

	/* copy fragment list */
	total = 0;
	mark  = DF_dnFragLast; /* last fragment */
	dcp = &dmacookie[frags - 1];
	tfp = &tfdp->frags[frags - 1];
	for (i = frags-1; i >= 0; dcp--, tfp--, i--, mark = 0) {
		total += dcp->dmac_size;
		tfp->DnFragLen  = LE32(dcp->dmac_size | mark);
		tfp->DnFragAddr = LE32(dcp->dmac_address);
	}

	tfdp->DnNextPtr = LE32(dp->tx_ring_dma +
				DN_LIST_SIZE * SLOT(slot + 1, TX_RING_SIZE));

	if ((flag & GEM_TXFLAG_INTR) != 0) {
		total |= FS_dnIndicate;
	}

#ifdef GEM_CONFIG_CKSUM_OFFLOAD
	if (IS_3C90XB_OR_LATER(lp)) {
		if ((flag & GEM_TXFLAG_TCP) != 0) {
			total |= FS_addTcpChecksum | FS_addIpChecksum;
		}
		else if ((flag & GEM_TXFLAG_UDP) != 0) {
			total |= FS_addUdpChecksum | FS_addIpChecksum;
		}
		else if ((flag & GEM_TXFLAG_IPv4) != 0) {
			total |= FS_addIpChecksum;
		}
	}
#endif
	ASSERT((total & FS_dnComplete) == 0);
	tfdp->FrameStartHeader = LE32(total);

	return (1);
}

static void
tcfe_tx_start(struct gem_dev *dp, int start_slot, int nslots)
{
	struct DPD_type0	*tfdp;
	int			prev_slot;
	uint32_t		head_dma;
	boolean_t		need_restart;
	struct tcfe_dev		*lp = (struct tcfe_dev *)dp->private;

	/* fix next pointer in the last descriptor of the new tx list */
	tfdp = (struct DPD_type0 *)(dp->tx_ring +
		 DN_LIST_SIZE * SLOT(start_slot + nslots - 1, TX_RING_SIZE));
	tfdp->DnNextPtr = 0;
	gem_tx_desc_dma_sync(dp, start_slot, nslots, DDI_DMA_SYNC_FORDEV);

	/* append the new tx descriptors at the tail of current tx list */
	prev_slot = SLOT(start_slot - 1, TX_RING_SIZE);
	tfdp = (struct DPD_type0 *) (dp->tx_ring + DN_LIST_SIZE * prev_slot);

	head_dma = ((uint32_t) dp->tx_ring_dma) + DN_LIST_SIZE * start_slot;
	tfdp->DnNextPtr = LE32(head_dma);
	gem_tx_desc_dma_sync(dp, prev_slot, 1, DDI_DMA_SYNC_FORDEV);

	if (IS_3C90X(lp) || !lp->dnlistptr_loaded) {
		/* stop the download engine */
		REG_LOCK(lp);
		EXEC_CMD_WAIT(dp, CmdDnStall);

		/* update DnListPtr if download engine has finished */
		if (INL(dp, WxDnListPtr) == 0) {
			OUTL(dp, WxDnListPtr, head_dma);
#ifdef TEST_BOOMERANG
			lp->dnlistptr_loaded = B_TRUE;
#endif
		}

		/* activate the download engine */
		EXEC_CMD(dp, CmdDnUnstall);
		REG_UNLOCK(lp);
	}
}

static void
tcfe_rx_desc_write(struct gem_dev *dp, int slot,
	    ddi_dma_cookie_t *dmacookie, int frags)
{
	int			i;
	struct UPD		*rfdp;
	struct UpFrag		*rfp;
	ddi_dma_cookie_t	*dcp;
	uint32_t		len;
	uint32_t		n;
	struct tcfe_dev		*lp = (struct tcfe_dev *)dp->private;

	rfdp = (struct UPD *) (dp->rx_ring + UP_LIST_SIZE * slot);

#if DEBUG_LEVEL > 2
	cmn_err(CE_CONT, "!%s: %s seqnum: %d, slot %d, frags: %d",
		dp->name, __func__, dp->rx_active_tail, slot, frags);
	for (i = 0; i < frags; i++) {
		cmn_err(CE_CONT, "  frag: %d addr: 0x%x, len: 0x%x",
			i, dmacookie[i].dmac_address, dmacookie[i].dmac_size);
	}
#endif

	/* copy fragment list */
	len = 0;
	dcp = dmacookie;
	rfp = rfdp->frags;
	for (i = frags - 1; i--; dcp++, rfp++) {
		len += n = (uint32_t) dcp->dmac_size;
		rfp->UpFragLen  = LE32(n);
		rfp->UpFragAddr = LE32((uint32_t) dcp->dmac_address);
	}

	/* for last fragment */
	len += n = (uint32_t) dcp->dmac_size;
	rfp->UpFragLen  = LE32(n | UF_upFragLast);
	rfp->UpFragAddr = LE32((uint32_t) dcp->dmac_address);

	/* make an upload header */
	/* XXX - this storing action must be the last for auto polling */
	if (lp->rx_last_slot != slot) {
		rfdp->UpPktStatus = 0;
	}
}

static void
tcfe_rx_start(struct gem_dev *dp, int start_slot, int nslots)
{
	int		i;
	struct UPD	*rfdp;
	struct tcfe_dev	*lp = (struct tcfe_dev *) dp->private;

	if (nslots > 1) {
		gem_tx_desc_dma_sync(dp,
			SLOT(start_slot + 1, RX_RING_SIZE), nslots - 1,
			DDI_DMA_SYNC_FORDEV);
	}

	/* enable the first descriptor of the new list */
	rfdp = (struct UPD *) (dp->rx_ring + UP_LIST_SIZE * start_slot);
	lp->rx_last_slot = SLOT(start_slot + nslots, RX_RING_SIZE);

	rfdp->UpPktStatus = 0;
	gem_tx_desc_dma_sync(dp, start_slot, 1, DDI_DMA_SYNC_FORDEV);

#ifndef TEST_BOOMERANG
	if (IS_3C90X(lp))
#endif
	{
		REG_LOCK(lp);
#ifdef never
		/* XXX - don't stall upload engine. it drops packets */
		EXEC_CMD_WAIT(dp, CmdUpStall);
#endif
		i = 0;
		while ((INW(dp, WxIntStatus) & IS_cmdInProgress) != 0) {
			if (i++ > 200) {
				cmn_err(CE_WARN,
					"%s: %s: timeout error: upLoad busy",
					dp->name, __func__);
				lp->need_to_reset = B_TRUE;
				break;
			}
			drv_usecwait(10);
		}

		/* start the stopped upload engine */
		EXEC_CMD(dp, CmdUpUnstall);
		REG_UNLOCK(lp);
	}
}

static uint_t
tcfe_tx_desc_stat(struct gem_dev *dp, int slot, int ndesc)
{
	int			ret;
	struct DPD_type0	*tfdp;
	uint32_t		header;
	struct tcfe_dev		*lp = (struct tcfe_dev *)dp->private;

#ifndef TEST_BOOMERANG
	if (IS_3C90XB_OR_LATER(lp)) {
		tfdp = (struct DPD_type0 *) (dp->tx_ring + DN_LIST_SIZE * slot);
		header = LE32(tfdp->FrameStartHeader);
		return ((header & FS_dnComplete) != 0 ? GEM_TX_DONE : 0);
	}
#endif
	ret = 0;
	REG_LOCK(lp);
	EXEC_CMD_WAIT(dp, CmdDnStall);
	if (INL(dp, WxDnListPtr) != 
		((uint32_t) dp->tx_ring_dma) + DN_LIST_SIZE * slot) {
		ret = GEM_TX_DONE;
	}
	EXEC_CMD(dp, CmdDnUnstall);
	REG_UNLOCK(lp);

	return (ret);
}

static uint64_t
tcfe_rx_desc_stat(struct gem_dev *dp, int slot, int ndesc)
{
	struct UPD	*rfdp;
	uint32_t	rfs;
	uint64_t	len;
	uint64_t	flag = 0;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	rfdp = (struct UPD *) (dp->rx_ring + UP_LIST_SIZE * slot);

	rfs  = LE32(rfdp->UpPktStatus);

	DPRINTF(2, (CE_CONT,
		"!%s: %s: slot:%d, rfs:0x%b",
		dp->name, __func__, slot, rfs, UpPktStatusBits));

	if ((rfs & UPS_upComplete) == 0) {
		/* not received */
		return (flag);
	}

	flag = GEM_RX_DONE | (rfs & UPS_upPktLen);

#define	UPS_upErrorVlan	\
	(UPS_alignmentError | UPS_upOverflow | UPS_runtFrame | \
	 UPS_crcError | UPS_upOverrun)

	if ((rfs & UPS_upErrorVlan) != 0) {
		/* error packet */

		DPRINTF(4, (CE_CONT,
			"!%s: %s: slot:%d, rfs:0x%b",
			dp->name, __func__, slot, rfs, UpPktStatusBits));

		dp->stats.errrcv++;
		if ((rfs & UPS_alignmentError) != 0) {
			dp->stats.frame++;
		}
		if ((rfs & UPS_upOverflow) != 0) {
			dp->stats.overflow++;
		}
		if ((rfs & UPS_upOverrun) != 0) {
			dp->stats.overflow++;
		}
		if ((rfs & UPS_runtFrame) != 0) {
			dp->stats.runt++;
		}
		if ((rfs & UPS_crcError) != 0) {
			dp->stats.crc++;
		}
		if ((rfs & UPS_oversizedFrame) != 0) {
			dp->stats.frame_too_long++;
		}

		flag = GEM_RX_ERR;
	}

#ifdef GEM_CONFIG_CKSUM_OFFLOAD
	else if ((rfs & (UPS_tcpChecksumChecked | UPS_tcpChecksumError))
						== UPS_tcpChecksumChecked) {
		flag |= GEM_RX_CKSUM_TCP;
	}
	else if ((rfs & (UPS_udpChecksumChecked | UPS_udpChecksumError))
						== UPS_udpChecksumChecked) {
		flag |= GEM_RX_CKSUM_UDP;
	}
	else if ((rfs & (UPS_ipChecksumChecked | UPS_ipChecksumError))
						== UPS_ipChecksumChecked) {
		flag |= GEM_RX_CKSUM_IPv4;
	}
#endif

	return (flag);
}

static void
tcfe_tx_desc_init(struct gem_dev *dp, int slot)
{
	struct DPD_type0	*tfdp;

	tfdp = (struct DPD_type0 *) (dp->tx_ring + DN_LIST_SIZE * slot);
	bzero(tfdp, DN_LIST_SIZE);
}

static void
tcfe_rx_desc_init(struct gem_dev *dp, int slot)
{
	struct UPD	*rfdp;
	uint32_t	next_dma;

	next_dma = ((uint32_t) dp->rx_ring_dma) +
				UP_LIST_SIZE * SLOT(slot + 1, RX_RING_SIZE);
	rfdp = (struct UPD *) (dp->rx_ring + UP_LIST_SIZE * slot);

	rfdp->UpPktStatus = LE32(UPS_upComplete);
	rfdp->UpNextPtr = LE32(next_dma);
}

static void
tcfe_tx_desc_clean(struct gem_dev *dp, int slot)
{
	struct DPD_type0	*tfdp;

	tfdp = (struct DPD_type0 *) (dp->tx_ring + DN_LIST_SIZE * slot);
	bzero(tfdp, DN_LIST_SIZE);
}

static void
tcfe_rx_desc_clean(struct gem_dev *dp, int slot)
{
	struct UPD	*rfdp;

	rfdp = (struct UPD *) (dp->rx_ring + UP_LIST_SIZE * slot);

	rfdp->UpPktStatus = LE32(UPS_upComplete);
}

/*
 * Device depend interrupt handler
 */
static void
tcfe_get_tx_status(struct gem_dev  *dp)
{
	int		error = 0;
	uint8_t		tx_status;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	DPRINTF(2, (CE_CONT, "%s: %s: called", dp->name, __func__));

	/* Clear Tx status stack. */
	while ((INW(dp, WxIntStatus) & IS_txComplete) != 0) {

		tx_status = INB(dp, WxTxStatus);
		OUTB(dp, WxTxStatus, 0);

		if ((tx_status & TS_txStatusOverflow) != 0) {
			cmn_err(CE_WARN,
			    "%s: tx status overflow: tx_status: %b",
			    dp->name, tx_status, TxStatusBits);
		}

		/* update stats */
		if ((tx_status & TS_txUnderrun) != 0) {
			/* fifo underflow */
			dp->stats.underflow++;
			dp->stats.errxmt++;
		}

		if ((tx_status & TS_maxCollisions) != 0) {
			/* exceed maximum collisions */
			dp->stats.excoll++;
			dp->stats.collisions += 16;
			dp->stats.errxmt++;
		}

		if ((tx_status & TS_txJabber) != 0) {
			dp->stats.errxmt++;
		}

		/* collect status bits */
		error |= (tx_status &
			     (TS_txJabber | TS_txUnderrun | TS_maxCollisions));
	}

	if (error != 0) {
		REG_LOCK(lp);
		if ((error & (TS_txUnderrun | TS_txJabber)) != 0) {
			/*
			 * XXX - not implemented completely
			 */
			EXEC_CMD_WAIT(dp, CmdTxReset);
			lp->need_to_reset = B_TRUE;
		}

		if ((error & TS_maxCollisions) != 0) {
			EXEC_CMD(dp, CmdStatisticsEnable);
		}

		EXEC_CMD(dp, CmdTxEnable); 
		REG_UNLOCK(lp);
	}
}

static u_int
tcfe_interrupt(struct gem_dev *dp)
{
	uint16_t	status;
	u_int		tx_sched = 0;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	/* Read interrupt status */
	status = INW(dp, WxIntStatus);
	DPRINTF(2, (CE_CONT, "%s: %s, status: %b",
		   dp->name, __func__, status, IntStatusBits));

	if ((status & OUR_INTR_BITS) == 0) {
		/* Not for us */
		return (DDI_INTR_UNCLAIMED);
	}

	DPRINTF(2, (CE_CONT, "%s: %s, status: %b",
		   dp->name, __func__, status, IntStatusBits));

	/*
	 * Clear all interrupt mask and ack interrupt,
	 */
	REG_LOCK(lp);
	EXEC_CMD(dp, CmdSetInterruptEnable | 0);
	EXEC_CMD(dp, CmdAcknowledgeInterrupt | (status & AllAck));
	REG_UNLOCK(lp);

	if (!dp->mac_active) {
		/*
		 * the device is not active.
		 * side effect: left interrupt masked.
		 */
		return (DDI_INTR_CLAIMED);
	}

	status &= OUR_INTR_BITS;

#ifdef GEM_CONFIG_POLLING
	/*
	 * XXX - the driver become funny when we use Countdown timer.
	 */
	if (dp->poll_interval != lp->last_poll_interval) {
		if (dp->poll_interval != 0) {
			/* polling mode */
			lp->intenable &= ~IS_upComplete;
		}
		else {
			/* normal mode */
			lp->intenable |= IS_upComplete;
		}

		/*
		 * To schedule the next timer interrupt,
		 * we pretend as we were interrupted from
		 * polling timer
		 */
		status |= IS_intRequested;

		lp->last_poll_interval = dp->poll_interval;
	}

	if ((status & IS_intRequested) != 0) {
		/*
		 * Reset timer  counter to schedule the next polling
		 * timer interrupt.
		 */
		OUTW(dp, WxCountdown, dp->poll_interval * 3);

		/* force to process upload complete  */
		status |= IS_upComplete;
	}
#endif /* GEM_CONFIG_POLLING */

	if ((status & IS_rxEarly) != 0) {
		cmn_err(CE_WARN, "%s: unexpected RX_EARLY, status: %b",
			dp->name, status, IntStatusBits);
	}

	if ((status & IS_upComplete) != 0) {
		/* Receive Complete */
		gem_receive(dp);
	}

	if ((status & IS_dnComplete) != 0) {
		if (gem_tx_done(dp)) {
			tx_sched = INTR_RESTART_TX;
		}
	}

	if ((status & IS_txComplete) != 0) {
		tcfe_get_tx_status(dp);
	}
	
	if ((status & IS_updateStats) != 0) {
		/* Update Statistics */
		tcfe_get_stats(dp);
	}

	if ((status & IS_hostError) != 0) {
		lp->need_to_reset = B_TRUE;
	}

	/*
	 * Postamble for interrupt thread
	 */
x:
	if (lp->need_to_reset) {
		gem_restart_nic(dp, B_TRUE);
		tx_sched = INTR_RESTART_TX;
	}

	/*
	 * Recover Interrput Enable register
	 */
	DPRINTF(4, (CE_CONT, "!%s: %s done: status: %b",
		dp->name, __func__, INW(dp, WxIntStatus), IntStatusBits));

	if ((dp->misc_flag & GEM_NOINTR) == 0) {
		/* enable interrupt again */
		REG_LOCK(lp);
		EXEC_CMD(dp, CmdSetInterruptEnable | lp->intenable);
		REG_UNLOCK(lp);
	}

	return (DDI_INTR_CLAIMED | tx_sched);
}
/*
 * HW depend MII routine
 */
#define MDIO_DELAY(dp)    {INW(dp, W4PhysicalMgmt);INW(dp, W4PhysicalMgmt);}
static void
tcfe_mii_sync(struct gem_dev *dp)
{
	int		i;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	/* output 32 ones */
	REG_LOCK(lp);
	SET_WIN(dp, 4);
	for (i = 0; i < 32; i++) {
		OUTW(dp, W4PhysicalMgmt, PM_mgmtDir | PM_mgmtData);
		MDIO_DELAY(dp);
		OUTW(dp, W4PhysicalMgmt,
				PM_mgmtDir | PM_mgmtData | PM_mgmtClk);
		MDIO_DELAY(dp);
	}
	REG_UNLOCK(lp);
}

static uint16_t
tcfe_mii_read(struct gem_dev *dp, uint_t reg)
{
	uint32_t	cmd;
	uint16_t	ret;
	int		i;
	uint16_t	data;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	cmd = MII_READ_CMD(dp->mii_phy_addr, reg);

	REG_LOCK(lp);
	SET_WIN(dp, 4);

	for (i = 31; i >= 18; i--) {
		data = ((cmd >> i) & 1) <<  PM_mgmtDataShift;
		OUTW(dp, W4PhysicalMgmt, data | PM_mgmtDir);
		MDIO_DELAY(dp);
		OUTW(dp, W4PhysicalMgmt, data | PM_mgmtDir | PM_mgmtClk);
		MDIO_DELAY(dp);
	}
	/* turn around */
	OUTW(dp, W4PhysicalMgmt, 0);
	MDIO_DELAY(dp);

	/* get response from PHY */
	OUTW(dp, W4PhysicalMgmt, PM_mgmtClk);
	MDIO_DELAY(dp);
	OUTW(dp, W4PhysicalMgmt, 0);
	if ((INW(dp, W4PhysicalMgmt) & PM_mgmtData) != 0) {
		DPRINTF(5, (CE_CONT, "%s: no response from phy@%d",
			dp->name, dp->mii_phy_addr));
	}

	OUTW(dp, W4PhysicalMgmt, PM_mgmtClk);
	MDIO_DELAY(dp);

	for (i = 16; i > 0; i--) {
		OUTW(dp, W4PhysicalMgmt, 0);
		ret = (ret << 1) |
			((INW(dp, W4PhysicalMgmt) >> PM_mgmtDataShift) & 1);
		OUTW(dp, W4PhysicalMgmt, PM_mgmtClk);
		MDIO_DELAY(dp);
	}

	/* send 2 ones to phy */
	for (i = 0; i < 2; i++) {
		OUTW(dp, W4PhysicalMgmt, PM_mgmtDir | PM_mgmtData);
		MDIO_DELAY(dp);
		OUTW(dp, W4PhysicalMgmt, PM_mgmtDir | PM_mgmtData | PM_mgmtClk);
		MDIO_DELAY(dp);
	}
	REG_UNLOCK(lp);
#ifndef notdef
	if (reg == MII_STATUS) {
		if ((ret & MII_STATUS_100_BASEX) != 0) {
			ret |= MII_STATUS_100_BASEX_FD;
		}
		if ((ret & MII_STATUS_10) != 0) {
			ret |= MII_STATUS_10_FD;
		}
	}
#endif
	return (ret);
}

static void
tcfe_mii_write(struct gem_dev *dp, uint_t reg, uint16_t val)
{
	int		i;
	uint32_t	cmd;
	uint16_t	data;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	cmd = MII_WRITE_CMD(dp->mii_phy_addr, reg, val);

	REG_LOCK(lp);
	SET_WIN(dp, 4);
	for (i = 31; i >= 0; i--) {
		data = ((cmd >> i) & 1) << PM_mgmtDataShift;
		OUTW(dp, W4PhysicalMgmt, data | PM_mgmtDir);
		MDIO_DELAY(dp);
		OUTW(dp, W4PhysicalMgmt, data | PM_mgmtDir | PM_mgmtClk);
		MDIO_DELAY(dp);
	}

	/* send 2 ones to phy */
	for (i = 0; i < 2; i++) {
		OUTW(dp, W4PhysicalMgmt, PM_mgmtDir | PM_mgmtData);
		MDIO_DELAY(dp);
		OUTW(dp, W4PhysicalMgmt, PM_mgmtDir | PM_mgmtData | PM_mgmtClk);
		MDIO_DELAY(dp);
	}
	REG_UNLOCK(lp);
}
#undef MDIO_DELAY

static void
tcfe_nomii_sync(struct gem_dev *dp)
{
	/* do nothing */
}

static uint16_t
tcfe_nomii_read(struct gem_dev *dp, uint_t index)
{
	uint16_t	ret;
	uint16_t	val;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	ret = 0;
	switch (index) {
	case MII_CONTROL:
		ret = lp->bmcr;
		break;

	case MII_STATUS:
		/* TODO: should return proper link status */
		/* setup capability */
		switch (lp->media) {
		case xcvrSelect100BASETX:
		case xcvrSelect100BASEFX:
			lp->bmsr = MII_STATUS_100_BASEX |
				   MII_STATUS_100_BASEX_FD;
			break;

		case xcvrSelect10BASET:
			lp->bmsr = MII_STATUS_10 | MII_STATUS_10_FD;
			break;

		default:
			ASSERT(lp->media != xcvrSelectMII);
			ASSERT(lp->media != xcvrSelectMIIExt);
			ASSERT(lp->media != xcvrSelectAN);
			lp->bmsr = MII_STATUS_10;
			break;
		}

		ret = lp->bmsr & ~MII_STATUS_LINKUP;

		switch (lp->media) {
		case xcvrSelectAUI:
		case xcvrSelect10BASE2:
			/* no way to detect link up */
			ret |= MII_STATUS_LINKUP;
			break;

		default:
			REG_LOCK(lp);
			SET_WIN(dp, 4);
			val = INW(dp, W4MediaStatus);
			if ((val & MS_linkDetect) != 0) {
				ret |= MII_STATUS_LINKUP;
			}
			REG_UNLOCK(lp);
			break;
		}

		lp->bmsr = ret;
		break;

	case MII_AN_ADVERT:
		ret = lp->adv;
		break;
	}
	return (ret);
}

static void
tcfe_nomii_write(struct gem_dev *dp, uint_t index, uint16_t val)
{
	struct tcfe_dev    *lp = (struct tcfe_dev *)dp->private;

	switch (index) {
	case MII_CONTROL:
		lp->bmcr = val & ~MII_CONTROL_RESET;
		break;

	case MII_AN_ADVERT:
		lp->adv = val;
		break;

	default:
		cmn_err(CE_WARN,
		"!%s: %s: writing to register %d in phy isn't permitted.",
			dp->name, __func__, index);
		break;
	}
}

static char *tcfe_media_name[] = {
	"10BASET",		/* 0 */
	"AUI",			/* 1 */
	"(reserved)",		/* 2 */
	"10BASE2",		/* 3 */
	"100BASETX",		/* 4 */
	"100BASEFX",		/* 5 */
	"MII",			/* 6 */
	"(reserved)",		/* 7 */
	"10/100Mbps w/ AN",	/* 8 */
	"external MII",		/* 9 */
	"(reserved)",		/* 10 */
	"(reserved)",		/* 11 */
	"(reserved)",		/* 12 */
	"(reserved)",		/* 13 */
	"(reserved)",		/* 14 */
	"(reserved)",		/* 15 */
};

static int
tcfe_mii_init(struct gem_dev *dp)
{
	uint32_t	ic;
	uint16_t	mopt;
	uint16_t	ms;
	uint32_t	port;
	struct tcfe_dev	*lp = (struct tcfe_dev *) dp->private;

	/*
	 * select media here
	 */

	/* read internal config register */
	REG_LOCK(lp);

	/*
	 * choose media port
	 */
	SET_WIN(dp, 3);
	mopt = INW(dp, W3MediaOptions);
	cmn_err(CE_CONT, "!%s: media options:%b",
		dp->name, mopt, 
		IS_3C90X(lp) ? MediaOptionsBits : MediaOptionsBits_BC);

	ic = INL(dp, W3InternalConfig);

	DPRINTF(0, (CE_CONT, "!%s: %s: internal_config:%b",
		dp->name, __func__, ic, InternalConfigBits));

	if ((ic & IC_autoSelect) != 0) {
		/* scan available media */
		if ((mopt & MO_baseT4Available) != 0) {
			lp->media = xcvrSelectMII;
			/* XXX - should we use 100M full forced mode ? */
		}
		else if (IS_3C90XB_OR_LATER(lp) &&
			(mopt & (MO_baseTxAvailable | MO_10bTAvailable)) != 0) {
			lp->media = xcvrSelectAN;
		}
		else if ((mopt & MO_baseTxAvailable) != 0) {
			lp->media = xcvrSelect100BASETX;
		}
		else if ((mopt & MO_baseFxAvailable) != 0) {
			lp->media = xcvrSelect100BASEFX;
		}
		else if ((mopt & MO_10BaseFL) != 0 && IS_3C90XB(lp)) {
			lp->media = xcvrSelectAUI;
		}
		else if ((mopt & MO_10bTAvailable) != 0) {
			lp->media = xcvrSelect10BASET;
		}
		else if ((mopt & MO_miiDevice) != 0) {
			lp->media = xcvrSelectMII;
		}
		else {
			lp->media = xcvrSelect10BASET;
		}

		/* fix tranceiver selection */
		ic = (ic & ~IC_xcvrSelect) |
				(lp->media << IC_xcvrSelectShift);
	}
	else {
		lp->media = (ic & IC_xcvrSelect) >> IC_xcvrSelectShift;
#ifdef NEVER
		if (IS_CARDBUS(lp)) {
			lp->media = xcvrSelectAN;
			ic = (ic & ~IC_xcvrSelect) |
					(lp->media << IC_xcvrSelectShift);
		}
#endif
	}
	OUTL(dp, W3InternalConfig, ic);

	cmn_err(CE_CONT, "!%s: media %s[%d] selected",
		dp->name,
		tcfe_media_name[lp->media], lp->media);

	/*
	 * setup media
	 */
	SET_WIN(dp, 4);
	ms = INW(dp, W4MediaStatus);
	ms &= ~MS_enableSqeStats;
	ms &= ~MS_jabberGuardEnable;
	ms &= ~MS_linkBeatEnale;

	switch (lp->media) {
	case xcvrSelectMII:
	case xcvrSelectAN:
		/* use MII with auto negotiation capability */
		dp->mii_phy_addr = 0;
		if (lp->media == xcvrSelectAN) {
			dp->mii_phy_addr = 24;
		}
		break;

	case xcvrSelect10BASET:
		dp->speed = GEM_SPD_10;
		ms |= MS_jabberGuardEnable | MS_linkBeatEnale;
		goto non_mii;

	case xcvrSelectAUI:
		dp->speed = GEM_SPD_10;
		ms |= MS_enableSqeStats;
		goto non_mii;

	case xcvrSelect10BASE2:
		dp->speed = GEM_SPD_10;
		goto non_mii;

	case xcvrSelect100BASETX:
	case xcvrSelect100BASEFX:
		dp->speed = GEM_SPD_100;
		ms |= MS_linkBeatEnale;
		goto non_mii;

	non_mii:
		dp->mii_fixedmode = B_TRUE;
		dp->gc.gc_mii_sync  = &tcfe_nomii_sync;
		dp->gc.gc_mii_read  = &tcfe_nomii_read;
		dp->gc.gc_mii_write = &tcfe_nomii_write;

		/* no need to scan phy */
		dp->mii_phy_addr = -1;
		break;

	default:
	case xcvrSelectMIIExt:
		REG_UNLOCK(lp);
		cmn_err(CE_WARN,
			"!%s: unknow media in InternalConfig: media:%d",
			dp->name, lp->media);
		return (GEM_FAILURE);
	}
	SET_WIN(dp, 4);
	OUTW(dp, W4MediaStatus, ms);

	REG_UNLOCK(lp);

	if (gem_mii_init_default(dp) == GEM_SUCCESS) {
#if DEBUG_LEVEL > 4
		int	i;
		for (i = 0; i < 32; i++) {
			DPRINTF(0, (CE_CONT, "%s: mii reg%d:%04x",
				dp->name, i, gem_mii_read(dp, i)));
		}
#endif
		return (GEM_SUCCESS);
	}
#ifdef notdef
	if (IS_3C90XB(lp)) {
		/* check built-in 10baseT port */
		REG_LOCK(lp);
		SET_WIN(dp, 3);
		val = INL(dp, W3InternalConfig);
		val = (val & ~IC_xcvrSelect) | IC_xcvrSelectAN;
		OUTL(dp, W3InternalConfig, val);
		REG_UNLOCK(lp);

		DPRINTF(0, (CE_CONT, "!%s: %s: internal_config:%b",
			dp->name, __func__, val, InternalConfigBits));

		/* search phy again */
		dp->mii_phy_addr = 0;
		if (gem_mii_init_default(dp) == GEM_SUCCESS) {
			return (GEM_SUCCESS);
		}
	}
#endif
	return (GEM_FAILURE);
}


static int
tcfe_read_eeprom(struct gem_dev *dp, uint_t offset)
{
	int		i;
	uint16_t	eecmd;
	uint16_t	val;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	/* window should be write-locked */
	eecmd = offset;

	if ((lp->chip->flags & EEADDRSIZE_8) != 0) {
		eecmd |= (EE_ReadRegister << (8 - 2));
	} else {
		eecmd |= (EE_ReadRegister << (6 - 2));
	}

	REG_LOCK(lp);
	SET_WIN(dp, 0);
	OUTW(dp, W0EepromCommand, eecmd);
	drv_usecwait(162);

	for (i = 0; (INW(dp, W0EepromCommand) & EC_eepromBusy) != 0; i++) {
		if (i > 20) {
			REG_UNLOCK(lp);
			cmn_err(CE_WARN, "!%s: %s: timeout",
				dp->name, __func__);
			return (-1);
		}
		drv_usecwait(10);
	}

	val = INW(dp, W0EepromData);
	REG_UNLOCK(lp);

	return (val);
}

#ifdef DEBUG_LEVEL
static void
tcfe_eeprom_dump(struct gem_dev *dp)
{
	int		i;
	uint16_t	*prom;
	uint16_t	sum;
	int		epsize;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

	epsize = 0x21;
	if (IS_CARDBUS(lp)) {
		epsize = 0x80;
	}
	/* allocate temporary buffer for reading eeprom contents */
	prom = (uint16_t *) kmem_alloc(epsize * sizeof(uint16_t), KM_SLEEP);

	for (i = 0; i < epsize; i++) {
		prom[i] = tcfe_read_eeprom(dp, i);
	}
	/* calculate checksum for eeprom area */
	sum = 0;
	for (i = 0; i < 0x20; i++) {
		sum += prom[i + lp->chip->ee_offset];
	}

	cmn_err(CE_CONT, "!%s: eeprom dump:", dp->name);
	for (i = 0; i < epsize; i += 4) {
		cmn_err(CE_CONT, "!0x%02x: 0x%04x 0x%04x 0x%04x 0x%04x", 
			i, prom[i], prom[i + 1], prom[i + 2], prom[i + 3]);
	}
	cmn_err(CE_CONT, "!sum[00-1f]: 0x%04x", sum);

	kmem_free(prom, epsize * sizeof(uint16_t));
}
#endif /* DEBUG_LEVEL */

#ifdef notdef
#include "tcfe_eeprom_wr.c"
#endif

static int
tcfe_attach_chip(struct gem_dev *dp)	
{
	int		i;
	int		val;
	uint8_t		*mac;
	struct tcfe_dev	*lp = (struct tcfe_dev *)dp->private;

#if DEBUG_LEVEL > 0
	tcfe_eeprom_dump(dp);
#endif

#ifdef notdef
	/* fix eeprom contents */
	val = tcfe_read_eeprom(dp,
			EO_SoftwareIntormation + lp->chip->ee_offset);

	tcfe_exec_eeprom_cmd(dp, EE_WriteEnable, 0, 0);
	tcfe_exec_eeprom_cmd(dp, EE_EraseRegister,
			EO_SoftwareIntormation + lp->chip->ee_offset, 0);

	tcfe_exec_eeprom_cmd(dp, EE_WriteEnable, 0, 0);
	tcfe_exec_eeprom_cmd(dp, EE_WriteRegister,
			EO_SoftwareIntormation + lp->chip->ee_offset,
			val | 0x8000);

	tcfe_exec_eeprom_cmd(dp, EE_WriteDisable, 0, 0);
#if DEBUG_LEVEL > 0
	tcfe_eeprom_dump(dp);
#endif
#endif

	mac = &dp->dev_addr.ether_addr_octet[0];
	for (i = 0; i < ETHERADDRL; i += 2) {
		val = tcfe_read_eeprom(dp,
				EO_OEMNodeAddress + i/2 + lp->chip->ee_offset);
		if (val < 0) {
			return (GEM_FAILURE);
		}
		/* the station address is stored in network byteorder */
		mac[i]     = (uint8_t) (val >> 8);
		mac[i + 1] = (uint8_t) val;
	}
#ifdef notdef
	gem_get_mac_addr_conf(dp);
#endif
	dp->txmaxdma = max(dp->txmaxdma, 512);
	dp->rxmaxdma = max(dp->rxmaxdma, 512);

#ifdef GEM_CONFIG_CKSUM_OFFLOAD
	if (IS_3C90XB_OR_LATER(lp)) {
		dp->misc_flag |= (GEM_CKSUM_FULL_IPv4 | GEM_CKSUM_HEADER_IPv4);
	} 
#endif
#ifdef CONFIG_VLAN
	if (IS_3C90XB_OR_LATER(lp)) {
		dp->misc_flag |= GEM_VLAN_HARD;
	} else {
		dp->misc_flag |= GEM_VLAN_SOFT;
	}
#endif
	return (GEM_SUCCESS);
}

/* ======================================================== */
/*
 * OS depend (device driver) routine
 */
/* ======================================================== */
static int
tcfeattach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int			i;
	int			n;
	ddi_iblock_cookie_t	c;
	ddi_acc_handle_t	conf_handle;
	uint16_t		comm;
	int			ret;
	int			vid;
	int			did;
	int			unit;
	struct chip_info	*p;
	int			val;
	int			len;
	struct pci_phys_spec	*regs;
	const char		*drv_name;
	struct gem_dev		*dp;
	struct tcfe_dev		*lp;
	void			*base;
	void			*base2;
	ddi_acc_handle_t	regs_handle;
	ddi_acc_handle_t	regs2_handle;
	struct gem_conf		*gcp;
	uint8_t			revid;
	uint8_t			lat;
	uint32_t		iline;

	unit =  ddi_get_instance(dip);
	drv_name = ddi_driver_name(dip);

	DPRINTF(3, (CE_CONT, "%s%d: tcfeattach: called", drv_name, unit));

	/*
	 * Check if chip is supported.
	 */
	if (pci_config_setup(dip, &conf_handle) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "%s: ddi_regs_map_setup failed",
			drv_name);
		goto err;
	}
	vid  = pci_config_get16(conf_handle, PCI_CONF_VENID);
	did  = pci_config_get16(conf_handle, PCI_CONF_DEVID);
	revid= pci_config_get8(conf_handle, PCI_CONF_REVID);
	comm = pci_config_get16(conf_handle, PCI_CONF_COMM);
	lat  = pci_config_get8(conf_handle, PCI_CONF_LATENCY_TIMER);
	iline  = pci_config_get32(conf_handle, PCI_CONF_ILINE);
	comm |= PCI_COMM_IO | PCI_COMM_ME | PCI_COMM_MAE;
	pci_config_put16(conf_handle, PCI_CONF_COMM, comm);

	pci_config_teardown(&conf_handle);

	switch (cmd) {
	case DDI_RESUME:
		return (gem_resume(dip));

	case DDI_ATTACH:
		for (i = 0, p = tcfe_chiptbl; i < CHIPTABLESIZE; i++, p++) {
			if (p->venid == vid && p->devid == did) {
				/* found */
				cmn_err(CE_CONT,
"!%s%d: %s (vid: 0x%04x, did: 0x%04x, revid: 0x%02x, latency timer: 0x%02x)",
				drv_name, unit, p->name, vid, did, revid, lat);
				goto chip_found;
			}
		}

		/* Not found */
		cmn_err(CE_WARN,
			"%s: tcfe_attach: wrong PCI venid/devid (0x%x, 0x%x)",
			drv_name, vid, did);
		p--;
chip_found:
		DPRINTF(0, (CE_CONT,
			"%s: tcfe_attach: iline %08x", drv_name, iline));
		/*
		 * Map in the device registers.
		 */
		if (gem_pci_regs_map_setup(dip,
			PCI_ADDR_IO, PCI_ADDR_MASK,
			&tcfe_dev_attr,
			(caddr_t *)&base, &regs_handle) != DDI_SUCCESS) {
			goto err;
		}

		regs2_handle = NULL;
		if ((p->flags & CHIP_CARDBUS) != 0) {
			if (gem_pci_regs_map_setup(dip,
				PCI_CONF_BASE2, PCI_REG_REG_M,
				&tcfe_dev_attr,
				(caddr_t *)&base2, &regs2_handle)
							!= DDI_SUCCESS) {
				ddi_regs_map_free(&regs_handle);
				goto err;
			}
		}

		lp = kmem_zalloc(sizeof(struct tcfe_dev), KM_SLEEP);
		lp->rev_id = revid;
		lp->chip   = p;
		lp->base2  = base2;
		lp->base2_ha = regs2_handle;

		/*
		 * construct gem configration
		 */
		gcp = (struct gem_conf *) kmem_zalloc(sizeof(*gcp), KM_SLEEP);

		/* name */
		sprintf(gcp->gc_name, "%s%d", drv_name, unit);

		/* consistency on tx and rx */
		gcp->gc_tx_buf_align   = sizeof(uint8_t) - 1;
#if 0
		/*
		 * XXX - 3c90x seems to hang on rx in multi tx fragment
		 * configuration.
		 */
		gcp->gc_tx_max_frags   =
				IS_3C90XB_OR_LATER(lp) ? MAXTXFRAGS : 1;
#else
		gcp->gc_tx_max_frags   = MAXTXFRAGS;
#endif
		gcp->gc_tx_max_descs_per_pkt = 1;
		gcp->gc_tx_desc_unit_shift = DN_LIST_SIZE_SHIFT; /* 64 byte */
		gcp->gc_tx_buf_size    = TX_BUF_SIZE;
		gcp->gc_tx_buf_limit   = gcp->gc_tx_buf_size - 1;
		gcp->gc_tx_ring_size   = TX_RING_SIZE;
		gcp->gc_tx_ring_limit  = min(TX_BUF_SIZE, TX_RING_SIZE - 1);
		gcp->gc_tx_auto_pad    = B_FALSE;
		gcp->gc_tx_copy_thresh = tcfe_tx_copy_thresh;
#ifdef CONFIG_OO
		gcp->gc_tx_desc_write_oo = TX_BUF_SIZE == TX_RING_SIZE;
#else
		gcp->gc_tx_desc_write_oo = B_FALSE;
#endif
		gcp->gc_rx_buf_align   = sizeof(uint8_t) - 1;
		gcp->gc_rx_max_frags   = MAXRXFRAGS;
		gcp->gc_rx_desc_unit_shift = UP_LIST_SIZE_SHIFT; /* 32 byte */
		gcp->gc_rx_ring_size   = RX_RING_SIZE;
		gcp->gc_rx_buf_max     = RX_RING_SIZE - 1;
		gcp->gc_rx_copy_thresh = tcfe_rx_copy_thresh;
		gcp->gc_rx_header_len  = 0;

		gcp->gc_io_area_size = 0;

		/* map attributes */
		gcp->gc_dev_attr = tcfe_dev_attr;
		gcp->gc_buf_attr = tcfe_buf_attr;
		gcp->gc_desc_attr = tcfe_buf_attr;

		/* dma attributes */
		gcp->gc_dma_attr_desc = tcfe_dma_attr_desc;
		gcp->gc_dma_attr_txbuf = tcfe_dma_attr_buf;
		gcp->gc_dma_attr_txbuf.dma_attr_sgllen = gcp->gc_tx_max_frags;
		gcp->gc_dma_attr_rxbuf = tcfe_dma_attr_buf;
		gcp->gc_dma_attr_rxbuf.dma_attr_sgllen = gcp->gc_rx_max_frags;

		/* time out parameters */
		gcp->gc_tx_timeout = 3*ONESEC;
		gcp->gc_tx_timeout_interval = ONESEC;

		/* flow control */
		gcp->gc_flow_control = FLOW_CONTROL_SYMMETRIC;

		/* MII timeout parameters */
		gcp->gc_mii_link_watch_interval = ONESEC;
		gcp->gc_mii_an_watch_interval   = ONESEC/5;
		gcp->gc_mii_reset_timeout = MII_RESET_TIMEOUT;	/* 1 sec */
		gcp->gc_mii_an_timeout = MII_AN_TIMEOUT;	/* 5 sec */
		gcp->gc_mii_an_wait = 0;
		gcp->gc_mii_linkdown_timeout = MII_LINKDOWN_TIMEOUT;

		/* MII work arounds */
		gcp->gc_mii_addr_min = 0;
		gcp->gc_mii_an_delay = 0;
		gcp->gc_mii_linkdown_action = MII_ACTION_RSA;
		gcp->gc_mii_linkdown_timeout_action = MII_ACTION_RESET;
		gcp->gc_mii_dont_reset = B_FALSE;

		/* I/O methods */

		/* mac operation */
		gcp->gc_attach_chip = &tcfe_attach_chip;
		gcp->gc_reset_chip  = &tcfe_reset_chip;
		gcp->gc_init_chip   = &tcfe_init_chip;
		gcp->gc_start_chip  = &tcfe_start_chip;
		gcp->gc_stop_chip   = &tcfe_stop_chip;
		gcp->gc_multicast_hash = &tcfe_mcast_hash;
		gcp->gc_set_rx_filter = &tcfe_set_rx_filter;
		gcp->gc_set_media   = &tcfe_set_media;
		gcp->gc_get_stats   = &tcfe_get_stats;
		gcp->gc_interrupt   = &tcfe_interrupt;

		/* descriptor operation */
		gcp->gc_tx_desc_write = &tcfe_tx_desc_write;
		gcp->gc_tx_start      = &tcfe_tx_start;
		gcp->gc_rx_desc_write = &tcfe_rx_desc_write;
		gcp->gc_rx_start      = &tcfe_rx_start;

		gcp->gc_tx_desc_init = &tcfe_tx_desc_init;
		gcp->gc_rx_desc_init = &tcfe_rx_desc_init;
		gcp->gc_tx_desc_stat = &tcfe_tx_desc_stat;
		gcp->gc_rx_desc_stat = &tcfe_rx_desc_stat;
		gcp->gc_tx_desc_clean = &tcfe_tx_desc_clean;
		gcp->gc_rx_desc_clean = &tcfe_rx_desc_clean;

		/* mii operations */
		gcp->gc_mii_init  = &tcfe_mii_init;
		gcp->gc_mii_config = &gem_mii_config_default;
		gcp->gc_mii_sync  = &tcfe_mii_sync;
		gcp->gc_mii_read  = &tcfe_mii_read;
		gcp->gc_mii_write = &tcfe_mii_write;
		gcp->gc_mii_tune_phy = NULL;

		mutex_init(&lp->reglock, NULL, MUTEX_DRIVER, NULL);
		dp = gem_do_attach(dip, gcp, base, &regs_handle,
					lp, sizeof(*lp));

		kmem_free(gcp, sizeof(*gcp));

		if (dp != NULL) {
			return (DDI_SUCCESS);
		}

		mutex_destroy(&lp->reglock);
err_free_mem:
		kmem_free(lp, sizeof(struct tcfe_dev));
err:
		return (DDI_FAILURE);
	}
	return (DDI_FAILURE);
}

static int
tcfedetach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int	ret;
	struct gem_dev  *dp;
	struct tcfe_dev   *lp;

	dp = GEM_GET_DEV(dip);
	lp = (struct tcfe_dev *) dp->private;

        switch (cmd) {
        case DDI_DETACH:
		if (lp->base2_ha != NULL) {
			ddi_regs_map_free(&lp->base2_ha);
		}
		mutex_destroy(&lp->reglock);
		return (gem_do_detach(dip));

	case DDI_SUSPEND:
		return (gem_suspend(dip));
	}

	return (DDI_FAILURE);
}

/* ======================================================== */
/*
 * OS depend (loadable streams driver) routine
 */
/* ======================================================== */
static	struct module_info tcfeminfo = {
	0,			/* mi_idnum */
	"tcfe",			/* mi_idname */
	0,			/* mi_minpsz */
	INFPSZ,			/* mi_maxpsz */
	64*1024,		/* mi_hiwat */
	1,			/* mi_lowat */
};

static	struct qinit tcferinit = {
	(int (*)()) NULL,	/* qi_putp */
	gem_rsrv,		/* qi_srvp */
	gem_open,		/* qi_qopen */
	gem_close,		/* qi_qclose */
	(int (*)()) NULL,	/* qi_qadmin */
	&tcfeminfo,		/* qi_minfo */
	NULL			/* qi_mstat */
};

static	struct qinit tcfewinit = {
	gem_wput,		/* qi_putp */
	gem_wsrv,		/* qi_srvp */
	(int (*)()) NULL,	/* qi_qopen */
	(int (*)()) NULL,	/* qi_qclose */
	(int (*)()) NULL,	/* qi_qadmin */
	&tcfeminfo,		/* qi_minfo */
	NULL			/* qi_mstat */
};

static struct streamtab	tcfe_info = {
	&tcferinit,	/* st_rdinit */
	&tcfewinit,	/* st_wrinit */
	NULL,		/* st_muxrinit */
	NULL		/* st_muxwrinit */
};

static	struct cb_ops cb_tcfe_ops = {
	nulldev,	/* cb_open */
	nulldev,	/* cb_close */
	nodev,		/* cb_strategy */
	nodev,		/* cb_print */
	nodev,		/* cb_dump */
	nodev,		/* cb_read */
	nodev,		/* cb_write */
	nodev,		/* cb_ioctl */
	nodev,		/* cb_devmap */
	nodev,		/* cb_mmap */
	nodev,		/* cb_segmap */
	nochpoll,	/* cb_chpoll */
	ddi_prop_op,	/* cb_prop_op */
	&tcfe_info,	/* cb_stream */
	D_NEW|D_MP	/* cb_flag */
};

static	struct dev_ops tcfe_ops = {
	DEVO_REV,	/* devo_rev */
	0,		/* devo_refcnt */
	gem_getinfo,	/* devo_getinfo */
	nulldev,	/* devo_identify */
	nulldev,	/* devo_probe */
	tcfeattach,	/* devo_attach */
	tcfedetach,	/* devo_detach */
	nodev,		/* devo_reset */
	&cb_tcfe_ops,	/* devo_cb_ops */
	NULL,		/* devo_bus_ops */
	ddi_power	/* devo_power */
};

static struct modldrv modldrv = {
	&mod_driverops,	/* Type of module.  This one is a driver */
	ident,
	&tcfe_ops,	/* driver ops */
};

static struct modlinkage modlinkage = {
	MODREV_1, &modldrv, NULL
};

/* ======================================================== */
/*
 * _init : done
 */
/* ======================================================== */
int
_init(void)
{
	int 	status;

	DPRINTF(2, (CE_CONT, "!tcfe: _init: called"));

	gem_mod_init(&tcfe_ops, "tcfe");
	status = mod_install(&modlinkage);
	if (status != DDI_SUCCESS) {
		gem_mod_fini(&tcfe_ops);
	}
	return (status);
}

/*
 * _fini : done
 */
int
_fini(void)
{
	int	status;

	DPRINTF(2, (CE_CONT, "!tcfe: _fini: called"));
	status = mod_remove(&modlinkage);
	if (status == DDI_SUCCESS) {
		gem_mod_fini(&tcfe_ops);
	}
	return (status);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}
